# coding: utf-8

import requests
import json
import re
from functools import lru_cache

from .idol import Idol
from .card import Card
from .infos import Event, Gacha, Info
from .enums import enum, blood_types, constellations, hands, home_towns, rarities
from .errors import CategoryNotFound, NotFound, NotValid

class KiraraException(Exception):
    def __init__(self, http_status, code, msg):
        self.http_status = http_status
        self.code = code
        self.msg = msg

    def __str__(self):
        return f"http status: {self.http_status} {self.code} - {self.msg}"

class Kirara(object):
    """A class that connects to the kirara api

    """
    max_retries = 10

    def __init__(self, requests_session=True, request_timeout=10):
        self.prefix = 'https://starlight.kirara.ca/api/v1/'
        self.request_timeout = request_timeout

        if isinstance(requests_session, requests.Session):
            self._session = requests_session
        else:
            if requests_session:
                self._session = requests.Session()
            else:
                from requests import api
                self._session = api

    def internal_call(self, method, url, payload, params):
        """Makes HTTP requests
        
        Parameters
        ----------
        method : str
            method of making HTTP requests ('GET', 'POST', etc.)

        url : str
            The url to make HTTP requests to

        payload : str
            Any additional parameters to pass

        params : dict
            Parameters to pass to a dict

        Returns
        ----------
        dict
            The result of the request in dict format
        """
        args = dict(params=params)
        args["timeout"] = self.request_timeout
        if not url.startswith('http'):
            url = self.prefix + url

        if payload:
            args["data"] = json.dumps(payload)

        r = self._session.request(method, url, **args)
        try:
            r.raise_for_status()
        finally:
            r.connection.close()
        if r.text and len(r.text) > 0 and r.text != 'null':
            results = r.json()
            return results
        else:
            return None

    @lru_cache(maxsize=None)
    def get(self, url, args=None, payload=None, **kwargs):
        """Make GET requests
        
        Parameters
        ----------
        url : str
            The url to the api (excluding the prefix url)

        args : dict
            Pass additional args

        payload : str
            Pass any additional parameters to request

        Returns
        ----------
        dict
            The result of the GET request
        """
        if args:
            kwargs.update(args)
        reconnect = self.max_retries
        while reconnect > 0:
            try:
                return self.internal_call('GET', url, payload, kwargs)
            except Exception as e:
                raise
                print('exception', str(e))

    def post(self, url, args=None, payload=None, **kwargs):
        """Make POST requests
        
        Parameters
        ----------
        url : str
            The url to the api (excluding the prefix url)

        args : dict
            Pass additional args

        payload : str
            Pass any additional parameters to request

        Returns
        ----------
        dict
            The result of the POST request
        """
        if args:
            kwargs.update(args)

        reconnect = self.max_retries
        while reconnect > 0:
            try:
                return self.internal_call('POST', url, payload, kwargs)
            except Exception as e:
                raise
                print('exception', str(e))

    def translate(self, translations: tuple):
        """Translate a tuple of Japanese strings to English

        Parameters
        ----------
        translate : tuple
            A tuple of eligible stings
            
        Returns
        ---------
        dict
            A dict of the translated text (JP:EN mapping)
        """
        results = self.post('read_tl', payload=translations)
        return results 

    def get_idol(self, idol_id: int):
        """Retrieve an idol's info

        Parameters
        ----------
        idol_id : int
            An Idol's ID to use from

        Returns
        ----------
        Idol
            An Idol object, which contains the idol's info
        """
        data = self.get("char_t/{0}".format(idol_id))
        if data['result'][0] is not None:

            return Idol(data['result'][0])
        else:
            raise NotFound("Idol ID can not be found in the Database. Is the ID correct?")

    def get_card(self, card_id: int, en_translate=False):
        """Retrieve a card's data
        
        Parameters
        ----------
        card_id : int
            A Card's ID to use from
        en_translate : bool
            Whether to translate the title, skill name, and lead skill to english
            
        Returns
        ----------
        Card
            A Card object, which contains the card's info
        """
        data = self.get("card_t/{0}".format(card_id))
        if data['result'][0] is not None:
            card = Card(data['result'][0])
            if en_translate:
                translations = (card.title, card.skill['skill_name'], card.lead_skill['name'])
                result = self.translate(translations)
                for strings, translated in result.items():
                    if translated is None:
                        translated = strings
                    
                card.title = card.title if result.get(card.title) is None else result.get(card.title)
                card.skill['skill_name'] = card.skill['skill_name'] if result.get(
                    card.skill['skill_name']) is None else result.get(card.skill['skill_name'])
                card.lead_skill['name'] = card.title if result.get(
                    card.lead_skill['name']) is None else result.get(card.lead_skill['name'])

                return card
            else:
                return card
        else:
            raise NotFound("Card ID can not be found in the Database. Is the ID correct?")
            
    def get_version(self):
        """Retrieve the client's version
        
        Returns
        ----------
        Info
            An Info object, which contains version info
        """
        data = self.get('info')

        return Info(data)

    def get_image(self, card: 'Card', category='card'):
        """Retrieve a Card's image data
        
        Parameters
        ----------
        card : Card
            A card object to use
        catergory : str
            What type of image to use (Default value is 'card')
            
        Returns
        ----------
        bytes
            The image bytes
        """
        if type(card) is Card:
            categories = {
                'card': card.image,
                'icon': card.icon,
                'spread': card.spread,
                'sprite': card.sprite
            }
            if category in categories:
                image = categories.get(category)

                response = self._session.get(image)

                return response
            else:
                raise CategoryNotFound("Defined Category not valid, use 'card', 'icon', 'spread', or 'sprite'")
        else:
            raise NotValid("The passed object for card is not a Card object")
        

    def get_now(self, category, en_translate=False):
        """Retrieve a list of occasions happenning in the game
        
        Parameters
        ----------
        category : str
            What type of event to iterate from
        en_translate : bool
            Whether to translate the event's name
            
        Returns
        --------
        list
            A list of gachas or events
        """
        categories = {
        'events': self.get("happening/now")['events'],
        'gachas': self.get("happening/now")['gachas']
        }
        happening_list = []

        if category in categories:
            stuff = categories.get(category)

            for index, event in enumerate(stuff):
                if category == 'events':
                    happening_list.append(Event(event))

                else:
                    for gacha, event in enumerate(stuff): # I don't know why you have to iterate again, maybe Im dumb

                        happening_list.append(Gacha(event))
            if en_translate:
                for thing in happening_list:
                    translations = self.translate((thing.name,)) # Tuples are weird

                    thing.name = translations.get(thing.name)
                
                return happening_list
            else:    
                return happening_list
                    
    def get_id(self, category, name, card_rarity=None, position=None):
        """Find a specific id based on parameters given
        
        Parameters
        ----------
        category : str
            Which category to search from ('card_t, or 'char_t')
        name : str
            An idol's name, full or just one part of a name
        card_rarity : str
            A rarity of card to look from (ranges from n to ssr, or n+ to ssr+)
        position : int
            Which card to get, based on release order
            
        Returns
        ----------
        list
            A list of cards matching the parameters
        int
            An ID of a specfic idol, or card"""

        cat_list = self.get('list/{0}'.format(category))['result']
        rarity = enum(rarities, card_rarity)  
        card_list= []

        for index, code in enumerate(cat_list):
            cat_name = cat_list[index]['conventional'].lower()
            match = bool(re.search(r"\b{0}\b".format(name), cat_name))

            if category == 'char_t':
                if match:
                    return int(cat_list[index]['chara_id'])

            elif category == 'card_t':
                if match:
                    if card_rarity:
                        if int(rarity) == cat_list[index]['rarity_dep']['rarity']:
                            card_list.append(int(cat_list[index]['id']))
                    else:
                        card_list.append(int(cat_list[index]['id']))
            else:
                raise CategoryNotFound("Category not found: Use 'card_t',or 'char_t'")
        if position:
            return card_list[position-1]
        else:
            return card_list
        



class BaseError(Exception):
    """Just a bare error"""

class CategoryNotFound(BaseError):
    """Defined Category not found"""
    pass

class NotFound(BaseError):
    """Something that couldn't be found"""
    pass

class NotValid(BaseError):
    """An Object that is not valid"""


class Info:
    """
    Represents the API's info

    Attributes
    ----------
    truth : str
        The game's (Deresute) truth version

    api_major : int
        The API's version major

    api_revision : int
        The API's revision number        

    """

    def __init__(self, info_data: dict):
        self.truth = info_data['truth_version']
        self._api_major = info_data['api_major']
        self._api_revision = info_data['api_revision']
        
class Gacha:
    """
    Represents gacha information

    Attributes
    ----------
    gacha : int
        A value that represents the gacha's position

    id : int
        The gacha's id

    name : str
        The gacha's name in japanese

    start_date : UNIX-datetime
        The gacha's start date

    end_date : UNIX-datetime    
        The gacha's end date

    type : int
        The gacha type

    subtype : int
        The gacha sub-type

    rates : dict
        The weighted rates for the gacha

    """
    def __init__(self, gacha_data: dict):
        self.id = gacha_data['id']
        self.name = gacha_data['name']
        self.start_date = gacha_data['start_date']
        self.end_date = gacha_data['end_date']
        self.type = gacha_data['type']
        self.subtype = gacha_data['subtype']
        self.rates = gacha_data['rates']

class Event:
    """
    Represents event info

    Attributes
    -------
    event : int
        Event index number in 'happening/now' endpoint

    id : int
        The event's id

    name : str
        Event name

    start_date : datetime obj
        Event start date

    end_date : datetime obj
        Event end date

    result_end_date : datetime obj
        The time left for the Event until it ends
    """

    def __init__(self, gacha_data):
        self.id = gacha_data['id']
        self.name = gacha_data['name']
        self.start_date = gacha_data['start_date']
        self.end_date = gacha_data['end_date']
        self.result_end_date = gacha_data['result_end_date']


import requests
from functools import lru_cache

from .enums import enum, attributes
from .errors import CategoryNotFound

class Card:
    """
    Represents a Card and its data

    Attributes
    ----------
    card_id : int
        The card's card id

    album_id : int
        The card's album id

    type : str
        The card's type (Cool, Cute, Passion, Office)

    image : str
        The card's image API link

    has_spread : bool
        True if the card has a spread image, otherwise false

    icon : str
        The card's icon link    

    chara_id : int
        The card's character id

    evo_id : int
        The card's evolution id of transformed card

    evo_type : int
        The card's evolution rarity type (From R to SSR)

    grow_type : int
        An integer that represents a boolean value, checks if card can grow or not

    name : str
        The card's name

    title : str
        The card's untranslated title

    open_dress_id : int
        The card's model id

    place : int
        The card's place value

    pose : int
        The card's sprite number

    series_id : int
        The evolution chain id of the card

    skill : dict
        The card's skill info in a dict

    lead_skill : dict
        The card's lead skill info in a dict

    rarity : dict
        Represents the card's rarity data in dict form

    min_vocal : int
        The card's minimum vocal value

    max_vocal : int
        The card's maximum vocal value

    bonus_vocal : int
        The card's bonus vocal value

    min_dance : int
        The card's minimum dance value

    max_dance : int
        The card's maximum dance value

    bonus_dance : int
        The card's bonus dance value

    min_visual : int
        The card's minimum visual value

    max_visual : int
        The card's maximum visual value

    bonus_visual : int
        The card's bonus visual value

    min_hp : int
        The card's minimum health value

    max_hp : int
        The card's maximum health value

    bonus_hp : int
        The card's bonus health value

    """
    def __init__(self, card_data: dict):
        self.card_id = card_data['id']
        self.album_id = card_data['album_id']
        self.type = card_data['attribute']

        self.image = card_data['card_image_ref']
        self.has_spread = card_data['has_spread']
        self.icon = card_data['icon_image_ref']
        self.sprite = card_data['sprite_image_ref']
        self.spread = card_data['spread_image_ref']
        self.sign = card_data['sign_image_ref']

        self.chara_id = card_data['chara_id']
        self.evo_id = card_data['evolution_id']
        self.evo_type = card_data['evolution_type']
        self.grow_type = card_data['grow_type']
        self.name = card_data['name']
        self.title = card_data['title']
        self.open_dress_id = card_data['open_dress_id']
        self.place = card_data['place']
        self.pose = card_data['pose']
        self.series_id = card_data['series_id']
        self.lead_skill = card_data['lead_skill']
        self.skill = card_data['skill']
        self.rarity = card_data['rarity']

        self.min_vocal = card_data['vocal_min']
        self.max_vocal = card_data['vocal_max']
        self.bonus_dance = card_data['bonus_dance']

        self.min_dance = card_data['dance_min']
        self.max_dance = card_data['dance_max']
        self.bonus_dance = card_data['bonus_dance']

        self.min_visual = card_data['visual_min']
        self.max_visual = card_data['visual_max']
        self.bonus_visual = card_data['bonus_visual']

        self.min_hp = card_data['hp_min']
        self.max_hp = card_data['hp_max']
        self.bonus_hp = card_data['bonus_hp']

    def min_max_stats(self, stat: str, level: int):
        """Calculates the value of a stat in a specific level
    
        Returns
        -------
        int
            Value of stat in specified level
        """
        if stat == 'dance':
            dance_formula = self.min_dance + (self.max_dance - self.min_dance) * (level/self.rarity['base_max_level'])

            return round(dance_formula)

        elif stat == 'visual':
            visual_formula = self.min_visual + (self.max_visual - self.min_visual) * (level/self.rarity['base_max_level'])

            return round(visual_formula)

        elif stat == 'vocal':
            vocal_formula = self.min_vocal + (self.max_vocal - self.min_vocal) * (level/self.rarity['base_max_level'])

            return round(vocal_formula)


from functools import lru_cache

from .enums import enum, blood_types, constellations, hands, home_towns

        
class Idol:
    """
    Represents an idol and her data

    Attributes
    ----------
    char_id : int
        The idol's character id

    age : int
        The idol's age

    bday : int
        The idol's birthday (not including month)

    bmonth : int
        The idol's birthmonth

    btype : str
        The idol's bloodtype

    bust : int
        The idol's bust measurement

    waist : int
        The idol's waist mearsurement

    hip : int
        The idol's hip measurement

    horoscope : str
        The idol's constellation horoscope

    conventional : str
        The idol's conventional name

    favorite : str
        The idol's favorite hobby

    hand : str
        The idol's dominant hand

    height : int
        The idol's height (in centimeters)

    home_town : str
        The idol's hometown

    kana_spaced : str
        The idol's name but in spaced kana

    kanji_spaced : str
        The idol's name but in kanji

    name : str
        The idol's name in japanese

    name_kana : str
        The idol's name in kana

    personality : int
        The idol's personality value

    type : str
        The idol's type (Cute, Cool, Passion, etc.)

    voice : str
        The idol's VA

    weight : int
        The idol's weight
    
    icon : str
        The link to the idol's icon
    """
    def __init__(self, char_data: dict):
        self.age = char_data['age']
        self.bday = char_data['birth_day']
        self.bmonth = char_data['birth_month']
        self.btype = enum(blood_types, char_data['blood_type'])
        self.bust = char_data['body_size_1']
        self.waist = char_data['body_size_2']
        self.hip = char_data['body_size_3']
        self.horoscope = enum(constellations, char_data['constellation'])
        self.conventional = char_data['conventional']
        self.favorite = char_data['favorite'] # NOTE: In Japanese due to read_tl not returning english
        self.hand = enum(hands, char_data['hand'])
        self.height = char_data['height']
        self.home_town = enum(home_towns, char_data['home_town'])
        self.kana_spaced = char_data['kana_spaced']
        self.kanji_spaced = char_data['kanji_spaced']
        self.name = char_data['name']
        self.name_kana = char_data['name_kana']
        self.personality = char_data['personality']
        self.type = char_data['type']
        self.voice = char_data['voice']
        self.weight = char_data['weight']
        self.icon = char_data['icon_image_ref']


import setuptools

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md") as f:
    readme = f.read()

setuptools.setup(name='pyKirara',
    author='EthanSk13s',
    url='https://github.com/EthanSk13s/pyKirara',
    project_urls={
    "Documentation": "https://pykirara.readthedocs.io/"
    },
    version='1.1.1',
    packages=['pyKirara'],
    license='MIT',
    description='a Python wrapper for the starlight.kirara API',
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    python_requires='>=3.5',
    classifiers=[
    'Development Status :: 5 - Production/Stable',
    'License :: OSI Approved :: MIT License',
    'Intended Audience :: Developers',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Internet',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Utilities',
    ]
)
